@@ -65,8 +65,8 @@ public class Main {
      * The program name and version.
      */
     public static final String VERSION = "(version 5.1)";
-    public static final String PROGRAM = "beagle.25Nov19.28d.jar";
-    public static final String COMMAND = "java -jar beagle.25Nov19.28d.jar";
+    public static final String PROGRAM = "beagle.18May20.d20.jar";
+    public static final String COMMAND = "java -jar beagle.18May20.d20.jar";
 
     /**
      * The copyright string.
@@ -78,7 +78,7 @@ public class Main {
      */
     public static final String SHORT_HELP = Main.PROGRAM + " " + VERSION
             + Const.nl + Main.COPYRIGHT
-            + Const.nl + "Enter \"java -jar beagle.25Nov19.28d.jar\" to "
+            + Const.nl + "Enter \"java -jar beagle.18May20.d20.jar\" to "
             + "list command line argument";
 
     private final Par par;
