@@ -23,10 +23,9 @@ import blbutil.Const;
 import blbutil.StringUtil;
 import blbutil.Utilities;
 import java.util.Arrays;
-import java.util.HashMap;
 import java.util.HashSet;
-import java.util.Map;
 import java.util.Set;
+import java.util.concurrent.ConcurrentHashMap;
 
 /**
  * <p>Class {@code BasicMarker} represents a genetic marker.
@@ -39,8 +38,8 @@ import java.util.Set;
 public class BasicMarker implements Marker {
 
     private static final String[] EMPTY_ID_ARRAY = new String[0];
-    private static final Map<String, String[]> ALLELES_MAP
-            = new HashMap<>(24);
+    private static final ConcurrentHashMap<String, String[]> ALLELES_MAP
+            = new ConcurrentHashMap<>(24); // allows re-use of String[] arrays
 
     private final int chromIndex;
     private final int pos;
@@ -189,7 +188,7 @@ public class BasicMarker implements Marker {
         }
     }
 
-    /* Return specified String[] alleles if not a SNV */
+    /* Returns specified String[] alleles if not a SNV */
     private static String[] canonicalAlleles(String[] alleles) {
         if (isSNV(alleles)) {
             String key = alleles[0];
