@@ -83,12 +83,13 @@ public final class FilterUtil {
      * more white-space characters
      */
     public static Filter<String> sampleFilter(File excludeSamplesFile) {
-        Filter<String> sampleFilter = null;
-        if (excludeSamplesFile!=null) {
-            Set<String> exclude = Utilities.idSet(excludeSamplesFile);
-            sampleFilter = Filter.excludeFilter(exclude);
+        if (excludeSamplesFile==null) {
+            return Filter.acceptAllFilter();
+        }
+        else {
+           Set<String> exclude = Utilities.idSet(excludeSamplesFile);
+           return Filter.excludeFilter(exclude);
         }
-        return sampleFilter;
     }
 
     /**
