@@ -18,23 +18,20 @@
  */
 package vcf;
 
-import blbutil.SampleFileIt;
-import beagleutil.Samples;
-import blbutil.Const;
 import blbutil.FileIt;
 import blbutil.Filter;
-import blbutil.Utilities;
+import blbutil.SampleFileIt;
 import java.io.File;
 import java.util.ArrayDeque;
+import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Deque;
 import java.util.List;
 import java.util.NoSuchElementException;
-import java.util.concurrent.ArrayBlockingQueue;
-import java.util.concurrent.BlockingQueue;
 import java.util.function.BiFunction;
 import java.util.function.Function;
 import java.util.stream.Collectors;
+import java.util.stream.IntStream;
 
 /**
  * <p>Class {@code VcfIt} represents  an iterator whose {@code next()}
@@ -49,69 +46,78 @@ import java.util.stream.Collectors;
  *
  * @author Brian L. Browning {@code <browning@uw.edu>}
  */
-public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
+public class VcfIt<E extends GTRec> implements SampleFileIt<E> {
 
     private static final float DEFAULT_MAX_LR = Float.MAX_VALUE;
 
     private final VcfHeader vcfHeader;
     private final FileIt<String> it;
     private final Function<String, E> mapper;
+    private String next;
     private final Filter<Marker> markerFilter;
-    private final Thread fileReaderThread;
-    private volatile boolean stopFileReadingThread = false;
 
-    private final BlockingQueue<String[]> stringBuffers;
-    private final Deque<E> emBuffer;
+    private final int bufferSize;
+    private final String[] stringBuffer;
+    private final Deque<E> recBuffer;
 
     /**
-     * The default number of VCF records stored in a buffer, which is 1000.
+     * The default number of VCF records stored in a buffer.
      */
-    public static final int MAX_EM_BUFFER_SIZE = 1000;
+    public static final int DEFAULT_BUFFER_SIZE = 1<<10;
 
-    /**
+   /**
      * A function mapping a string VCF record with GT format fields
      * to a {@code GTRec} object.
      */
     public static final BiFunction<VcfHeader, String, GTRec> toBitSetGT
-            = (VcfHeader h, String s) -> new BitSetGT(h, s);
+            = (VcfHeader h, String s) -> new BitSetGTRec(h, s);
 
     /**
      * A function mapping a string VCF record with GL format fields
      * to a {@code VcfRecord} object.
      */
     public static final BiFunction<VcfHeader, String, GTRec> toGLRec
-            = (VcfHeader h, String s) -> VcfRecord.fromGL(h, s, DEFAULT_MAX_LR);
+            = (VcfHeader h, String s) -> VcfRec.fromGL(h, s, DEFAULT_MAX_LR);
 
     /**
      * A function mapping a string VCF record with GT or GL format fields
      * to a {@code VcfRecord} object.
      */
     public static final BiFunction<VcfHeader, String, GTRec> toGTGLRec
-            = (VcfHeader h, String s) -> VcfRecord.fromGTGL(h, s, DEFAULT_MAX_LR);
+            = (VcfHeader h, String s) -> VcfRec.fromGTGL(h, s, DEFAULT_MAX_LR);
 
     /**
      * A function mapping a string VCF record with GT or GL format fields
      * to a {@code VcfRecord} object.
      */
-    public static final BiFunction<VcfHeader, String, VcfRecord> toVcfRecord
-            = (VcfHeader h, String s) -> VcfRecord.fromGTGL(h, s, DEFAULT_MAX_LR);
+    public static final BiFunction<VcfHeader, String, VcfRec> toVcfRecord
+            = (VcfHeader h, String s) -> VcfRec.fromGTGL(h, s, DEFAULT_MAX_LR);
 
     /**
-     * Create and returns a new {@code VcfIt} instance from the specified
-     * objects.
-     * @param <R> the type returned by the returned {@code VcfIt}
-     * @param strIt an iterator that returns lines of a VCF file
-     * @param recMapper a function mapping string VCF records to
-     * {@code GTRec} objects
-     * @return a new {@code VcfIt} instance
-     * @throws IllegalArgumentException if a format error is detected in a
-     * line of a VCF file returned by {@code strIt}
-     * @throws NullPointerException if
-     * {@code strIt == null || recMapper == null}
+     * Returns an array containing VCF meta-information lines, the
+     * VCF header line, and the first VCF data line.
+     * @param src a string describing the source of the VCF file
+     * @param it an iterator that returns the lines of a VCF file
+     * @return an array containing VCF meta-information lines, the
+     * VCF header line, and the first VCF data line
+     * @throws NullPointerException if {@code it == null}
+     * @throws IllegalArgumentException if all lines returned by the iterator
+     * begin with the '#' character
      */
-    public static <R extends GTRec> VcfIt<R> create(
-            FileIt<String> strIt, BiFunction<VcfHeader, String, R> recMapper) {
-        return VcfIt.create(strIt, Filter.acceptAllFilter(), recMapper);
+    static String[] head(String src, FileIt<String> it) {
+        String hash = "#";
+        List<String> lines = new ArrayList<>(32);
+        String line =  it.hasNext() ? it.next() : null;
+        while (line!=null && line.startsWith(hash)) {
+            lines.add(line);
+            line = it.hasNext() ? it.next() : null;
+        }
+        if (line==null) {
+            throw new IllegalArgumentException("ERROR: missing VCF data lines ("
+                    + src + ")");
+        }
+        lines.add(line);
+        return lines.toArray(new String[0]);
     }
 
     /**
@@ -119,20 +125,18 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
      * objects.
      * @param <R> the type returned by the returned {@code VcfIt}
      * @param strIt an iterator that returns lines of a VCF file
-     * @param sampleFilter a sample filter or {@code null}
      * @param recMapper a function mapping string VCF records to
      * {@code GTRec} objects
      * @return a new {@code VcfIt} instance
      * @throws IllegalArgumentException if a format error is detected in a
      * line of a VCF file returned by {@code strIt}
      * @throws NullPointerException if
-     * {@code strIt == null || recMapper == null}
+     * {@code strIt == null || mapFactory == null}
      */
     public static <R extends GTRec> VcfIt<R> create(
-            FileIt<String> strIt, Filter<String> sampleFilter,
-            BiFunction<VcfHeader, String, R> recMapper) {
-        return VcfIt.create(strIt, sampleFilter, Filter.acceptAllFilter(),
-                recMapper);
+            FileIt<String> strIt, BiFunction<VcfHeader, String, R> recMapper) {
+        return VcfIt.create(strIt, Filter.acceptAllFilter(),
+                Filter.acceptAllFilter(), recMapper);
     }
 
     /**
@@ -148,14 +152,14 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
      * @throws IllegalArgumentException if a format error is detected in a
      * line of a VCF file returned by {@code strIt}
      * @throws NullPointerException if
-     * {@code strIt == null || recMapper == null}
+     * {@code strIt == null || mapFactory == null}
      */
     public static <R extends GTRec> VcfIt<R> create(
             FileIt<String> strIt, Filter<String> sampleFilter,
             Filter<Marker> markerFilter,
             BiFunction<VcfHeader, String, R> recMapper) {
         return VcfIt.create(strIt, sampleFilter, markerFilter, recMapper,
-                MAX_EM_BUFFER_SIZE);
+                DEFAULT_BUFFER_SIZE);
     }
 
     /**
@@ -167,157 +171,83 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
      * @param markerFilter a marker filter or {@code null}
      * @param recMapper a function mapping string VCF records to
      * {@code GTRec} objects
-     * @param bufferSize the buffer size
+     * @param bufferSize the requested buffer size
      * @return a new {@code VcfIt} instance
      * @throws IllegalArgumentException if a format error is detected in a
      * line of a VCF file returned by {@code strIt}
      * @throws IllegalArgumentException if {@code bufferSize < 1}
      * @throws NullPointerException if
-     * {@code strIt == null || recMapper == null}
+     * {@code strIt == null || mapFactory == null}
      */
     public static <R extends GTRec> VcfIt<R> create(
             FileIt<String> strIt, Filter<String> sampleFilter,
             Filter<Marker> markerFilter,
             BiFunction<VcfHeader, String, R> recMapper, int bufferSize) {
-        VcfIt<R> vcfIt = new VcfIt<>(strIt, sampleFilter, markerFilter,
+        return new VcfIt<>(strIt, sampleFilter, markerFilter,
                 recMapper, bufferSize);
-        vcfIt.start();
-        return vcfIt;
     }
 
     private VcfIt(FileIt<String> it, Filter<String> sampleFilter,
             Filter<Marker> markerFilter,
-            BiFunction<VcfHeader, String, E> recMapper, int bufferSize) {
+            BiFunction<VcfHeader, String, E> recMapper,
+            int bufferSize) {
         if (bufferSize < 1) {
             throw new IllegalArgumentException(String.valueOf(bufferSize));
         }
         if (markerFilter==null) {
             markerFilter = Filter.acceptAllFilter();
         }
-        this.vcfHeader = new VcfHeader(it, sampleFilter);
+        String src = it.file()==null ? "stdin" : it.file().getName();
+        String[] head = head(src, it);
+        String[] nonDataLines = Arrays.copyOf(head, head.length-1);
+        String firstDataLine = head[head.length-1];
+        boolean[] isDiploid = VcfHeader.isDiploid(firstDataLine);
         this.it = it;
+        this.vcfHeader = new VcfHeader(src, nonDataLines, isDiploid, sampleFilter);
         this.mapper = (String s) -> recMapper.apply(vcfHeader, s);
+        this.next = firstDataLine;
         this.markerFilter = markerFilter;
-        this.stringBuffers = new ArrayBlockingQueue<>(1);
-        this.emBuffer = new ArrayDeque<>(bufferSize);
-        this.fileReaderThread = fileReadingThread();
-    }
-
-    private void start() {
-        this.fileReaderThread.setDaemon(true);
-        this.fileReaderThread.start();
+        this.bufferSize = bufferSize;
+        this.stringBuffer = new String[stringBufferSize(next, bufferSize)];
+        this.recBuffer = new ArrayDeque<>(bufferSize);
         fillEmissionBuffer();
-        if (emBuffer.isEmpty()) {
-            noRecordFoundError(it);
-        }
     }
 
-    private void noRecordFoundError(FileIt<String> it) {
-        if (it.hasNext()==false) {
-            StringBuilder sb = new StringBuilder(100);
-            sb.append("No VCF records found (data source: ");
-            sb.append(it.file()==null ? "stdin" : it.file());
-            sb.append(")");
-            sb.append(Const.nl);
-            sb.append("Check that the chromosome identifiers are the same in each input VCF");
-            sb.append(Const.nl);
-            sb.append("file and in the \'chrom=\' command line argument (if \'chrom=\' is used).");
-            throw new IllegalArgumentException(sb.toString());
-        }
-    }
-
-    private Thread fileReadingThread() {
-        Runnable runnable = () -> {
-            String line = readLine(it);
-            int bufferSize = stringBufferSize(line);
-            while (line != null && stopFileReadingThread == false) {
-                String chromPlusTab = chromFieldPlusTab(line);
-                String[] sa = new String[bufferSize];
-                int size = 0;
-                while (line != null && size < bufferSize
-                        && line.startsWith(chromPlusTab)) {
-                    sa[size++] = line;
-                    line = readLine(it);
-                }
-                if (size < bufferSize) {
-                    sa = Arrays.copyOf(sa, size);
-                }
-                putInBlockingQueue(stringBuffers, sa);
-            }
-            if (stopFileReadingThread == false) {
-                putInBlockingQueue(stringBuffers, new String[0]);    // sentinel
-            }
-        };
-        return new Thread(runnable);
-    }
-
-    private static int stringBufferSize(String line) {
-        if (line == null) {
-            return 0;
-        }
-        long nBytesPerLine = 2*line.length();
-        Runtime rt = Runtime.getRuntime();
-        long maxMem = rt.maxMemory();
-        if (maxMem == Long.MAX_VALUE) {
-            maxMem = 500 * (1 << 30);
-        }
-        long bufferSize = 1 + (maxMem / (20*nBytesPerLine));
-        if (bufferSize > MAX_EM_BUFFER_SIZE) {
-            bufferSize = MAX_EM_BUFFER_SIZE;
-        }
-        return (int) bufferSize;
-    }
-
-    private static <E> void putInBlockingQueue(BlockingQueue<E> q, E e) {
-        try {
-            q.put(e);
-        } catch (InterruptedException ex) {
-            Utilities.exit("ERROR: InterruptedException", ex);
-        }
-    }
-
-    private static <E> E takeFromBlockingQueue(BlockingQueue<E> q) {
-        try {
-            return q.take();
-        } catch (InterruptedException ex) {
-            Utilities.exit("ERROR: InterruptedException", ex);
-        }
-        assert false;
-        return null;
-    }
-
-    private static String chromFieldPlusTab(String vcfRecord) {
-        int tabIndex = vcfRecord.indexOf(Const.tab);
-        if (tabIndex == -1) {
-            String s = Const.nl + "ERROR: Missing tab delimiter in VCV Record:"
-                    + Const.nl + vcfRecord
-                    + Const.nl + "Exiting Program";
-            Utilities.exit(s);
-        }
-        return vcfRecord.substring(0, tabIndex + 1);
+    private static int stringBufferSize(String line, int maxBufferSize) {
+        long maxMem = Runtime.getRuntime().maxMemory();
+        long nBytesPerLine = 2*(line==null ? 0 : line.length());
+        long bufferSize = 1 + (maxMem >> 4)/(nBytesPerLine);
+        return (int) Math.min(bufferSize, maxBufferSize);
     }
 
     private void fillEmissionBuffer() {
-        assert emBuffer.isEmpty();
-        int lastLength = -1;
-        while (lastLength != 0 && emBuffer.size() < MAX_EM_BUFFER_SIZE) {
-            String[] stringBuffer = takeFromBlockingQueue(stringBuffers);
-            lastLength = stringBuffer.length;
-            if (stringBuffer.length>0) {
-                List<E> list = Arrays.stream(stringBuffer)
+        assert recBuffer.isEmpty();
+        int size = -1;
+        while (size!=0 && recBuffer.size()<bufferSize) {
+            size = fillStringBuffer(stringBuffer.length);
+            if (size>0) {
+                List<E> list = IntStream.range(0, size)
                         .parallel()
+                        .mapToObj(j -> stringBuffer[j])
                         .map(mapper)
                         .filter(e -> markerFilter.accept(e.marker()))
                         .collect(Collectors.toList());
-                emBuffer.addAll(list);
-            }
-            else {
-                // put sentinel element back
-                putInBlockingQueue(stringBuffers, stringBuffer);
+                recBuffer.addAll(list);
             }
         }
     }
 
+    private int fillStringBuffer(int maxSize) {
+        int size = 0;
+        if (next != null) {
+            while (next!=null && size<maxSize) {
+                stringBuffer[size++] = next;
+                next = readLine(it);
+            }
+        }
+        return size;
+    }
+
     private static String readLine(FileIt<String> it) {
         if (it.hasNext()==false) {
             return null;
@@ -331,15 +261,11 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
 
     @Override
     public void close() {
-        stopFileReadingThread = true;
-        stringBuffers.poll();  // unblock file reading thread
-        try {
-            fileReaderThread.join();
-        } catch (InterruptedException ex) {
-            Utilities.exit("Error: InterruptedException", ex);
-        }
+        next = null;
         it.close();
-        emBuffer.clear();
+        recBuffer.clear();
+        Arrays.fill(stringBuffer, null);
+
     }
 
     /**
@@ -349,7 +275,7 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
      */
     @Override
     public boolean hasNext() {
-        return !emBuffer.isEmpty();
+        return !recBuffer.isEmpty();
     }
 
     /**
@@ -362,8 +288,8 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
         if (!hasNext()) {
             throw new NoSuchElementException();
         }
-        E first = emBuffer.removeFirst();
-        if (emBuffer.isEmpty()) {
+        E first = recBuffer.removeFirst();
+        if (recBuffer.isEmpty()) {
             fillEmissionBuffer();
         }
         return first;
@@ -375,8 +301,7 @@ public class VcfIt<E extends MarkerContainer> implements SampleFileIt<E> {
      */
     @Override
     public void remove() {
-        String s = "remove() is not supported by VcfIterator";
-        throw new UnsupportedOperationException(s);
+        throw new UnsupportedOperationException(this.getClass().toString());
     }
 
     @Override
